import logging
from typing import Iterator, List, Optional, Tuple

from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.nodes import (
    SolcContractDefinition,
    SolcEnumDefinition,
    SolcErrorDefinition,
    SolcFunctionDefinition,
    SolcImportDirective,
    SolcPragmaDirective,
    SolcSourceUnit,
    SolcStructDefinition,
    SolcUserDefinedValueTypeDefinition,
    SolcUsingForDirective,
    SolcVariableDeclaration,
)

from ..declaration.abc import DeclarationAbc
from ..declaration.contract_definition import ContractDefinition
from ..declaration.enum_definition import EnumDefinition
from ..declaration.error_definition import ErrorDefinition
from ..declaration.function_definition import FunctionDefinition
from ..declaration.struct_definition import StructDefinition
from ..declaration.user_defined_value_type_definition import (
    UserDefinedValueTypeDefinition,
)
from ..declaration.variable_declaration import VariableDeclaration
from ..utils import IrInitTuple
from .import_directive import ImportDirective
from .pragma_directive import PragmaDirective
from .using_for_directive import UsingForDirective

logger = logging.getLogger(__name__)


class SourceUnit(SolidityAbc):
    """
    Source unit is the root node.

    !!! warning
        Source unit [byte_location][woke.ast.ir.abc.IrAbc.byte_location] does not cover the whole file.
        Only lines 3-7 are covered by the source unit in the following example:
        ```solidity linenums="1"
        // SPDX-License-Identifier: MIT

        pragma solidity ^0.8;

        contract Foo {
            function bar() public {}
        }
        ```
        Also trailing newlines are not covered by the source unit.
    """

    _ast_node: SolcSourceUnit

    _file_source: bytes
    _license: Optional[str]
    _source_unit_name: str
    _pragmas: List[PragmaDirective]
    _imports: List[ImportDirective]
    _declared_variables: List[VariableDeclaration]
    _enums: List[EnumDefinition]
    _functions: List[FunctionDefinition]
    _structs: List[StructDefinition]
    _errors: List[ErrorDefinition]
    _user_defined_value_types: List[UserDefinedValueTypeDefinition]
    _contracts: List[ContractDefinition]
    _using_for_directives: List[UsingForDirective]

    def __init__(
        self,
        init: IrInitTuple,
        source_unit: SolcSourceUnit,
    ):
        super().__init__(init, source_unit, None)
        self._file_source = init.source
        self._license = source_unit.license
        self._source_unit_name = source_unit.absolute_path

        self._pragmas = []
        self._imports = []
        self._declared_variables = []
        self._enums = []
        self._functions = []
        self._structs = []
        self._errors = []
        self._user_defined_value_types = []
        self._contracts = []
        self._using_for_directives = []
        for node in source_unit.nodes:
            if isinstance(node, SolcPragmaDirective):
                self._pragmas.append(PragmaDirective(init, node, self))
            elif isinstance(node, SolcImportDirective):
                self._imports.append(ImportDirective(init, node, self))
            elif isinstance(node, SolcVariableDeclaration):
                self._declared_variables.append(VariableDeclaration(init, node, self))
            elif isinstance(node, SolcEnumDefinition):
                self._enums.append(EnumDefinition(init, node, self))
            elif isinstance(node, SolcFunctionDefinition):
                self._functions.append(FunctionDefinition(init, node, self))
            elif isinstance(node, SolcStructDefinition):
                self._structs.append(StructDefinition(init, node, self))
            elif isinstance(node, SolcErrorDefinition):
                self._errors.append(ErrorDefinition(init, node, self))
            elif isinstance(node, SolcUserDefinedValueTypeDefinition):
                self._user_defined_value_types.append(
                    UserDefinedValueTypeDefinition(init, node, self)
                )
            elif isinstance(node, SolcContractDefinition):
                self._contracts.append(ContractDefinition(init, node, self))
            elif isinstance(node, SolcUsingForDirective):
                self._using_for_directives.append(UsingForDirective(init, node, self))
            else:
                assert False, f"Unknown node type: {node}"

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for pragma in self._pragmas:
            yield from pragma
        for import_directive in self._imports:
            yield from import_directive
        for variable_declaration in self._declared_variables:
            yield from variable_declaration
        for enum in self._enums:
            yield from enum
        for function in self._functions:
            yield from function
        for struct in self._structs:
            yield from struct
        for error in self._errors:
            yield from error
        for user_defined_value_type in self._user_defined_value_types:
            yield from user_defined_value_type
        for contract in self._contracts:
            yield from contract
        for using_for_directive in self._using_for_directives:
            yield from using_for_directive

    @property
    def parent(self) -> None:
        """
        Returns:
            Does not have a parent.
        """
        return None

    @property
    def file_source(self) -> bytes:
        """
        As opposed to [source][woke.ast.ir.abc.IrAbc.source], this property returns the whole file source.
        Returns:
            Source code of the file including trailing newlines and license string.
        """
        return self._file_source

    @property
    def license(self) -> Optional[str]:
        """
        !!! example
            Returns `MIT` for the following license comment:
            ```solidity
            // SPDX-License-Identifier: MIT
            ```
        Returns:
            License string of the file, if any.
        """
        return self._license

    @property
    def source_unit_name(self) -> str:
        """
        Returns:
            Source unit name of the file.
        """
        return self._source_unit_name

    @property
    def pragmas(self) -> Tuple[PragmaDirective, ...]:
        """
        Returns:
            Pragma directives present in the file.
        """
        return tuple(self._pragmas)

    @property
    def imports(self) -> Tuple[ImportDirective, ...]:
        """
        Returns:
            Import directives present in the file.
        """
        return tuple(self._imports)

    @property
    def declared_variables(self) -> Tuple[VariableDeclaration, ...]:
        """
        Should only return constants.
        Returns:
            Top-level variable declarations present in the file.
        """
        return tuple(self._declared_variables)

    @property
    def enums(self) -> Tuple[EnumDefinition, ...]:
        """
        Returns:
            Top-level enum definitions present in the file.
        """
        return tuple(self._enums)

    @property
    def functions(self) -> Tuple[FunctionDefinition, ...]:
        """
        Should only return [FunctionDefinitions][woke.ast.ir.declaration.function_definition.FunctionDefinition] of the [FunctionKind.FREE_FUNCTION][woke.ast.enums.FunctionKind.FREE_FUNCTION] kind.
        Returns:
            Top-level function definitions present in the file.
        """
        return tuple(self._functions)

    @property
    def structs(self) -> Tuple[StructDefinition, ...]:
        """
        Returns:
            Top-level struct definitions present in the file.
        """
        return tuple(self._structs)

    @property
    def errors(self) -> Tuple[ErrorDefinition, ...]:
        """
        Returns:
            Top-level error definitions present in the file.
        """
        return tuple(self._errors)

    @property
    def user_defined_value_types(self) -> Tuple[UserDefinedValueTypeDefinition, ...]:
        """
        Returns:
            Top-level user-defined value type definitions present in the file.
        """
        return tuple(self._user_defined_value_types)

    @property
    def contracts(self) -> Tuple[ContractDefinition, ...]:
        """
        Returns:
            Contract definitions present in the file.
        """
        return tuple(self._contracts)

    @property
    def using_for_directives(self) -> Tuple[UsingForDirective, ...]:
        """
        Returns:
            Top-level using for directives present in the file.
        """
        return tuple(self._using_for_directives)

    def declarations_iter(self) -> Iterator[DeclarationAbc]:
        """
        Yields:
            All declarations ([DeclarationAbc][woke.ast.ir.declaration.abc.DeclarationAbc]) present in the file (recursively).
        """
        yield from self.declared_variables
        yield from self.enums
        for enum in self.enums:
            yield from enum.values
        yield from self.functions
        yield from self.structs
        yield from self.errors
        yield from self.user_defined_value_types
        yield from self.contracts
        for contract in self.contracts:
            yield from contract.declarations_iter()



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from functools import lru_cache
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from ..declaration.variable_declaration import VariableDeclaration
    from ..declaration.user_defined_value_type_definition import (
        UserDefinedValueTypeDefinition,
    )
    from ..expression.elementary_type_name_expression import (
        ElementaryTypeNameExpression,
    )
    from ..expression.new_expression import NewExpression
    from ..meta.using_for_directive import UsingForDirective
    from .array_type_name import ArrayTypeName

from woke.ast.ir.abc import SolidityAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import (
    SolcArrayTypeName,
    SolcElementaryTypeName,
    SolcFunctionTypeName,
    SolcMapping,
    SolcTypeNameUnion,
    SolcUserDefinedTypeName,
    TypeDescriptionsModel,
)
from woke.ast.types import (
    Address,
    Array,
    Bool,
    Bytes,
    Contract,
    Enum,
    Fixed,
    FixedBytes,
    Function,
    Int,
    Mapping,
    String,
    Struct,
    Type,
    TypeAbc,
    UFixed,
    UInt,
    UserDefinedValueType,
)
from woke.utils.string import StringReader

logger = logging.getLogger(__name__)


class TypeNameAbc(SolidityAbc, ABC):
    """
    Abstract base class for all IR type name nodes.
    """

    _type_descriptions: TypeDescriptionsModel

    def __init__(
        self, init: IrInitTuple, type_name: SolcTypeNameUnion, parent: SolidityAbc
    ):
        super().__init__(init, type_name, parent)
        self._type_descriptions = type_name.type_descriptions

    @staticmethod
    def from_ast(
        init: IrInitTuple, type_name: SolcTypeNameUnion, parent: SolidityAbc
    ) -> TypeNameAbc:
        from .array_type_name import ArrayTypeName
        from .elementary_type_name import ElementaryTypeName
        from .function_type_name import FunctionTypeName
        from .mapping import Mapping
        from .user_defined_type_name import UserDefinedTypeName

        if isinstance(type_name, SolcArrayTypeName):
            return ArrayTypeName(init, type_name, parent)
        elif isinstance(type_name, SolcElementaryTypeName):
            return ElementaryTypeName(init, type_name, parent)
        elif isinstance(type_name, SolcFunctionTypeName):
            return FunctionTypeName(init, type_name, parent)
        elif isinstance(type_name, SolcMapping):
            return Mapping(init, type_name, parent)
        elif isinstance(type_name, SolcUserDefinedTypeName):
            return UserDefinedTypeName(init, type_name, parent)

    @property
    @abstractmethod
    def parent(
        self,
    ) -> Union[
        VariableDeclaration,
        UserDefinedValueTypeDefinition,
        ElementaryTypeNameExpression,
        NewExpression,
        UsingForDirective,
        ArrayTypeName,
        Mapping,
    ]:
        """
        Returns:
            Parent node of the type name.
        """
        ...

    @property
    @lru_cache(maxsize=2048)
    def type(
        self,
    ) -> Union[
        Array,
        Address,
        Bool,
        Int,
        UInt,
        Fixed,
        UFixed,
        String,
        Bytes,
        FixedBytes,
        Type,
        Function,
        Mapping,
        Struct,
        Enum,
        Contract,
        UserDefinedValueType,
    ]:
        """
        Returns:
            Type of the type name.
        """
        assert self._type_descriptions.type_identifier is not None

        type_identifier = StringReader(self._type_descriptions.type_identifier)
        ret = TypeAbc.from_type_identifier(
            type_identifier, self._reference_resolver, self.cu_hash
        )
        assert (
            len(type_identifier) == 0 and ret is not None
        ), f"Failed to parse type identifier: {self._type_descriptions.type_identifier}"
        assert isinstance(
            ret,
            (
                Array,
                Address,
                Bool,
                Int,
                UInt,
                Fixed,
                UFixed,
                String,
                Bytes,
                FixedBytes,
                Type,
                Function,
                Mapping,
                Struct,
                Enum,
                Contract,
                UserDefinedValueType,
            ),
        )
        return ret

    @property
    def type_string(self) -> str:
        """
        !!! example
            `:::solidity mapping(uint256 => int24[])` in the case of the `:::solidity mapping(uint => int24[])` type name in the following declaration:
            ```solidity
            mapping(uint => int24[]) map;
            ```

        Returns:
            User-friendly string describing the type name type.
        """
        assert self._type_descriptions.type_string is not None
        return self._type_descriptions.type_string



from functools import lru_cache, partial
from typing import Iterator, Optional, Set, Tuple

from woke.ast.enums import ModifiesStateFlag, UnaryOpOperator
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.declaration.function_definition import FunctionDefinition
from woke.ast.ir.expression.abc import ExpressionAbc
from woke.ast.ir.reference_resolver import CallbackParams
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import AstNodeId, SolcUnaryOperation


class UnaryOperation(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcUnaryOperation
    _parent: SolidityAbc

    _operator: UnaryOpOperator
    _prefix: bool
    _sub_expression: ExpressionAbc
    _function_id: Optional[AstNodeId]

    def __init__(
        self,
        init: IrInitTuple,
        unary_operation: SolcUnaryOperation,
        parent: SolidityAbc,
    ):
        super().__init__(init, unary_operation, parent)
        self._operator = unary_operation.operator
        self._prefix = unary_operation.prefix
        self._sub_expression = ExpressionAbc.from_ast(
            init, unary_operation.sub_expression, self
        )
        self._function_id = unary_operation.function
        if self._function_id is not None:
            init.reference_resolver.register_post_process_callback(self._post_process)

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._sub_expression

    def _post_process(self, callback_params: CallbackParams):
        function = self.function
        assert function is not None
        function.register_reference(self)
        self._reference_resolver.register_destroy_callback(
            self.file, partial(self._destroy, function)
        )

    def _destroy(self, function: FunctionDefinition) -> None:
        function.unregister_reference(self)

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def operator(self) -> UnaryOpOperator:
        return self._operator

    @property
    def prefix(self) -> bool:
        return self._prefix

    @property
    def sub_expression(self) -> ExpressionAbc:
        return self._sub_expression

    @property
    def is_ref_to_state_variable(self) -> bool:
        return False

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        ret = self.sub_expression.modifies_state

        if (
            self.operator
            in {
                UnaryOpOperator.PLUS_PLUS,
                UnaryOpOperator.MINUS_MINUS,
                UnaryOpOperator.DELETE,
            }
            and self.sub_expression.is_ref_to_state_variable
        ):
            ret |= {(self, ModifiesStateFlag.MODIFIES_STATE_VAR)}
        return ret

    @property
    def function(self) -> Optional[FunctionDefinition]:
        if self._function_id is None:
            return None
        node = self._reference_resolver.resolve_node(self._function_id, self._cu_hash)
        assert isinstance(node, FunctionDefinition)
        return node



import logging
from pathlib import Path
from typing import List, Optional, Tuple, Union

import woke.ast.ir.yul as yul
from woke.ast.ir.abc import IrAbc
from woke.ast.ir.declaration.variable_declaration import VariableDeclaration
from woke.ast.ir.expression.identifier import Identifier
from woke.ast.ir.expression.member_access import MemberAccess
from woke.ast.ir.meta.identifier_path import IdentifierPath
from woke.ast.ir.type_name.user_defined_type_name import UserDefinedTypeName
from woke.lsp.common_structures import (
    DocumentUri,
    Location,
    LocationLink,
    PartialResultParams,
    Position,
    StaticRegistrationOptions,
    TextDocumentPositionParams,
    TextDocumentRegistrationOptions,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from woke.lsp.context import LspContext
from woke.lsp.utils import path_to_uri, position_within_range, uri_to_path
from woke.lsp.utils.position import changes_to_byte_offset

logger = logging.getLogger(__name__)


class TypeDefinitionOptions(WorkDoneProgressOptions):
    pass


class TypeDefinitionRegistrationOptions(
    TextDocumentRegistrationOptions, TypeDefinitionOptions, StaticRegistrationOptions
):
    pass


class TypeDefinitionParams(
    TextDocumentPositionParams, WorkDoneProgressParams, PartialResultParams
):
    pass


def _get_results_from_node(
    original_node: IrAbc,
    position: Position,
    context: LspContext,
    byte_offset: int,
    node_name_location: Optional[Tuple[int, int]],
) -> Optional[List[Tuple[Path, Tuple[int, int]]]]:
    if isinstance(original_node, VariableDeclaration):
        assert node_name_location is not None
        name_location_range = context.compiler.get_range_from_byte_offsets(
            original_node.file, node_name_location
        )
        if not position_within_range(position, name_location_range):
            return None

    if isinstance(original_node, (Identifier, MemberAccess)):
        node = original_node.referenced_declaration
        if node is None:
            return None
    elif isinstance(original_node, (IdentifierPath, UserDefinedTypeName)):
        part = original_node.identifier_path_part_at(byte_offset)
        if part is None:
            return None
        node = part.referenced_declaration
    elif isinstance(original_node, yul.Identifier):
        external_reference = original_node.external_reference
        if external_reference is None:
            return None
        node = external_reference.referenced_declaration
    else:
        node = original_node

    if not isinstance(node, VariableDeclaration):
        return None

    type_name = node.type_name
    if isinstance(type_name, UserDefinedTypeName):
        return [
            (
                type_name.referenced_declaration.file,
                type_name.referenced_declaration.name_location,
            )
        ]
    else:
        return None


async def _get_type_definition_from_cache(
    path: Path,
    position: Position,
    context: LspContext,
):
    new_byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, position.line, position.character
    )
    backward_changes = context.compiler.get_last_compilation_backward_changes(path)
    if backward_changes is None:
        raise Exception("No backward changes found for path")
    changes_before = backward_changes[0:new_byte_offset]
    old_byte_offset = changes_to_byte_offset(changes_before) + new_byte_offset

    tree = context.compiler.last_compilation_interval_trees[path]

    intervals = tree.at(old_byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    if len(nodes) == 0:
        raise ValueError(f"Could not find node at {old_byte_offset}")

    node = max(nodes, key=lambda n: n.ast_tree_depth)

    if isinstance(node, VariableDeclaration):
        location = node.name_location
        forward_changes = context.compiler.get_last_compilation_forward_changes(path)
        if forward_changes is None:
            raise Exception("No forward changes found for path")
        new_start = (
            changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
        )
        new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
        node_name_location = (new_start, new_end)
    else:
        node_name_location = None

    result = _get_results_from_node(
        node, position, context, old_byte_offset, node_name_location
    )
    if result is None:
        return None

    if len(result) == 1:
        path, location = result[0]
        forward_changes = context.compiler.get_last_compilation_forward_changes(path)
        if forward_changes is None:
            raise Exception("No forward changes found for path")
        new_start = (
            changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
        )
        new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
        return Location(
            uri=DocumentUri(path_to_uri(path)),
            range=context.compiler.get_range_from_byte_offsets(
                path, (new_start, new_end)
            ),
        )
    else:
        ret = []
        for path, location in result:
            forward_changes = context.compiler.get_last_compilation_forward_changes(
                path
            )
            if forward_changes is None:
                raise Exception("No forward changes found for path")
            new_start = (
                changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
            )
            new_end = (
                changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
            )
            ret.append(
                Location(
                    uri=DocumentUri(path_to_uri(path)),
                    range=context.compiler.get_range_from_byte_offsets(
                        path, (new_start, new_end)
                    ),
                )
            )
        return ret


async def type_definition(
    context: LspContext, params: TypeDefinitionParams
) -> Union[None, Location, List[Location], List[LocationLink]]:
    logger.debug(
        f"Type definition for file {params.text_document.uri} at position {params.position} requested"
    )

    path = uri_to_path(params.text_document.uri).resolve()
    if (
        path not in context.compiler.interval_trees
        or not context.compiler.output_ready.is_set()
    ):
        # try to use old build artifacts
        try:
            return await _get_type_definition_from_cache(path, params.position, context)
        except Exception:
            pass

    await context.compiler.output_ready.wait()
    if path not in context.compiler.interval_trees:
        return None

    tree = context.compiler.interval_trees[path]

    byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, params.position.line, params.position.character
    )
    intervals = tree.at(byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    logger.debug(
        f"Found {len(nodes)} nodes at byte offset {byte_offset}:\n{sorted(nodes, key=lambda x: x.ast_tree_depth)}"
    )
    if len(nodes) == 0:
        return None

    node = max(nodes, key=lambda n: n.ast_tree_depth)
    logger.debug(f"Found node {node}")

    if isinstance(node, VariableDeclaration):
        node_name_location = node.name_location
    else:
        node_name_location = None

    result = _get_results_from_node(
        node, params.position, context, byte_offset, node_name_location
    )
    if result is None:
        return None

    if len(result) == 1:
        path, location = result[0]
        return Location(
            uri=DocumentUri(path_to_uri(path)),
            range=context.compiler.get_range_from_byte_offsets(path, location),
        )
    else:
        return [
            Location(
                uri=DocumentUri(path_to_uri(path)),
                range=context.compiler.get_range_from_byte_offsets(path, location),
            )
            for path, location in result
        ]

    if isinstance(node, VariableDeclaration):
        name_location_range = context.compiler.get_range_from_byte_offsets(
            node.file, node.name_location
        )
        if not position_within_range(params.position, name_location_range):
            return None

    if isinstance(node, (Identifier, MemberAccess)):
        node = node.referenced_declaration
        if node is None:
            return None
    elif isinstance(node, (IdentifierPath, UserDefinedTypeName)):
        part = node.identifier_path_part_at(byte_offset)
        if part is None:
            return None
        node = part.referenced_declaration
    elif isinstance(node, yul.Identifier):
        external_reference = node.external_reference
        if external_reference is None:
            return None
        node = external_reference.referenced_declaration

    if not isinstance(node, VariableDeclaration):
        return None

    type_name = node.type_name
    if isinstance(type_name, UserDefinedTypeName):
        return [
            Location(
                uri=DocumentUri(path_to_uri(type_name.referenced_declaration.file)),
                range=context.compiler.get_range_from_byte_offsets(
                    type_name.referenced_declaration.file,
                    type_name.referenced_declaration.name_location,
                ),
            )
        ]
    else:
        return None



from __future__ import annotations

import logging
import re
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import nullcontext
from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable, DefaultDict, Dict, List, Optional, Tuple, Type, Union

import rich.console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.syntax import Syntax, SyntaxTheme
from rich.tree import Tree

import woke.ast.ir.yul as yul
import woke.cli.console
from woke.ast.ir.abc import IrAbc
from woke.ast.ir.declaration.contract_definition import ContractDefinition
from woke.ast.ir.declaration.enum_definition import EnumDefinition
from woke.ast.ir.declaration.enum_value import EnumValue
from woke.ast.ir.declaration.error_definition import ErrorDefinition
from woke.ast.ir.declaration.event_definition import EventDefinition
from woke.ast.ir.declaration.function_definition import FunctionDefinition
from woke.ast.ir.declaration.modifier_definition import ModifierDefinition
from woke.ast.ir.declaration.struct_definition import StructDefinition
from woke.ast.ir.declaration.user_defined_value_type_definition import (
    UserDefinedValueTypeDefinition,
)
from woke.ast.ir.declaration.variable_declaration import VariableDeclaration
from woke.ast.ir.expression.assignment import Assignment
from woke.ast.ir.expression.binary_operation import BinaryOperation
from woke.ast.ir.expression.conditional import Conditional
from woke.ast.ir.expression.elementary_type_name_expression import (
    ElementaryTypeNameExpression,
)
from woke.ast.ir.expression.function_call import FunctionCall
from woke.ast.ir.expression.function_call_options import FunctionCallOptions
from woke.ast.ir.expression.identifier import Identifier
from woke.ast.ir.expression.index_access import IndexAccess
from woke.ast.ir.expression.index_range_access import IndexRangeAccess
from woke.ast.ir.expression.literal import Literal
from woke.ast.ir.expression.member_access import MemberAccess
from woke.ast.ir.expression.new_expression import NewExpression
from woke.ast.ir.expression.tuple_expression import TupleExpression
from woke.ast.ir.expression.unary_operation import UnaryOperation
from woke.ast.ir.meta.identifier_path import IdentifierPath
from woke.ast.ir.meta.import_directive import ImportDirective
from woke.ast.ir.meta.inheritance_specifier import InheritanceSpecifier
from woke.ast.ir.meta.modifier_invocation import ModifierInvocation
from woke.ast.ir.meta.override_specifier import OverrideSpecifier
from woke.ast.ir.meta.parameter_list import ParameterList
from woke.ast.ir.meta.pragma_directive import PragmaDirective
from woke.ast.ir.meta.source_unit import SourceUnit
from woke.ast.ir.meta.structured_documentation import StructuredDocumentation
from woke.ast.ir.meta.try_catch_clause import TryCatchClause
from woke.ast.ir.meta.using_for_directive import UsingForDirective
from woke.ast.ir.statement.block import Block
from woke.ast.ir.statement.break_statement import Break
from woke.ast.ir.statement.continue_statement import Continue
from woke.ast.ir.statement.do_while_statement import DoWhileStatement
from woke.ast.ir.statement.emit_statement import EmitStatement
from woke.ast.ir.statement.expression_statement import ExpressionStatement
from woke.ast.ir.statement.for_statement import ForStatement
from woke.ast.ir.statement.if_statement import IfStatement
from woke.ast.ir.statement.inline_assembly import InlineAssembly
from woke.ast.ir.statement.placeholder_statement import PlaceholderStatement
from woke.ast.ir.statement.return_statement import Return
from woke.ast.ir.statement.revert_statement import RevertStatement
from woke.ast.ir.statement.try_statement import TryStatement
from woke.ast.ir.statement.unchecked_block import UncheckedBlock
from woke.ast.ir.statement.variable_declaration_statement import (
    VariableDeclarationStatement,
)
from woke.ast.ir.statement.while_statement import WhileStatement
from woke.ast.ir.type_name.array_type_name import ArrayTypeName
from woke.ast.ir.type_name.elementary_type_name import ElementaryTypeName
from woke.ast.ir.type_name.function_type_name import FunctionTypeName
from woke.ast.ir.type_name.mapping import Mapping
from woke.ast.ir.type_name.user_defined_type_name import UserDefinedTypeName
from woke.config import WokeConfig
from woke.utils.file_utils import is_relative_to

logger = logging.getLogger(__name__)


detectors: List[Detector] = []

visit_map: Dict[Type[IrAbc], Callable] = {
    ContractDefinition: lambda self, node: self.visit_contract_definition(node),
    EnumDefinition: lambda self, node: self.visit_enum_definition(node),
    EnumValue: lambda self, node: self.visit_enum_value(node),
    ErrorDefinition: lambda self, node: self.visit_error_definition(node),
    EventDefinition: lambda self, node: self.visit_event_definition(node),
    FunctionDefinition: lambda self, node: self.visit_function_definition(node),
    ModifierDefinition: lambda self, node: self.visit_modifier_definition(node),
    StructDefinition: lambda self, node: self.visit_struct_definition(node),
    UserDefinedValueTypeDefinition: lambda self, node: self.visit_user_defined_value_type_definition(
        node
    ),
    VariableDeclaration: lambda self, node: self.visit_variable_declaration(node),
    Assignment: lambda self, node: self.visit_assignment(node),
    BinaryOperation: lambda self, node: self.visit_binary_operation(node),
    Conditional: lambda self, node: self.visit_conditional(node),
    ElementaryTypeNameExpression: lambda self, node: self.visit_elementary_type_name_expression(
        node
    ),
    FunctionCall: lambda self, node: self.visit_function_call(node),
    FunctionCallOptions: lambda self, node: self.visit_function_call_options(node),
    Identifier: lambda self, node: self.visit_identifier(node),
    IndexAccess: lambda self, node: self.visit_index_access(node),
    IndexRangeAccess: lambda self, node: self.visit_index_range_access(node),
    Literal: lambda self, node: self.visit_literal(node),
    MemberAccess: lambda self, node: self.visit_member_access(node),
    NewExpression: lambda self, node: self.visit_new_expression(node),
    TupleExpression: lambda self, node: self.visit_tuple_expression(node),
    UnaryOperation: lambda self, node: self.visit_unary_operation(node),
    IdentifierPath: lambda self, node: self.visit_identifier_path(node),
    ImportDirective: lambda self, node: self.visit_import_directive(node),
    InheritanceSpecifier: lambda self, node: self.visit_inheritance_specifier(node),
    ModifierInvocation: lambda self, node: self.visit_modifier_invocation(node),
    OverrideSpecifier: lambda self, node: self.visit_override_specifier(node),
    ParameterList: lambda self, node: self.visit_parameter_list(node),
    PragmaDirective: lambda self, node: self.visit_pragma_directive(node),
    SourceUnit: lambda self, node: self.visit_source_unit(node),
    StructuredDocumentation: lambda self, node: self.visit_structured_documentation(
        node
    ),
    TryCatchClause: lambda self, node: self.visit_try_catch_clause(node),
    UsingForDirective: lambda self, node: self.visit_using_for_directive(node),
    Block: lambda self, node: self.visit_block(node),
    Break: lambda self, node: self.visit_break(node),
    Continue: lambda self, node: self.visit_continue(node),
    DoWhileStatement: lambda self, node: self.visit_do_while_statement(node),
    EmitStatement: lambda self, node: self.visit_emit_statement(node),
    ExpressionStatement: lambda self, node: self.visit_expression_statement(node),
    ForStatement: lambda self, node: self.visit_for_statement(node),
    IfStatement: lambda self, node: self.visit_if_statement(node),
    InlineAssembly: lambda self, node: self.visit_inline_assembly(node),
    PlaceholderStatement: lambda self, node: self.visit_placeholder_statement(node),
    Return: lambda self, node: self.visit_return(node),
    RevertStatement: lambda self, node: self.visit_revert_statement(node),
    TryStatement: lambda self, node: self.visit_try_statement(node),
    UncheckedBlock: lambda self, node: self.visit_unchecked_block(node),
    VariableDeclarationStatement: lambda self, node: self.visit_variable_declaration_statement(
        node
    ),
    WhileStatement: lambda self, node: self.visit_while_statement(node),
    ArrayTypeName: lambda self, node: self.visit_array_type_name(node),
    ElementaryTypeName: lambda self, node: self.visit_elementary_type_name(node),
    FunctionTypeName: lambda self, node: self.visit_function_type_name(node),
    Mapping: lambda self, node: self.visit_mapping(node),
    UserDefinedTypeName: lambda self, node: self.visit_user_defined_type_name(node),
    yul.Assignment: lambda self, node: self.visit_yul_assignment(node),
    yul.Block: lambda self, node: self.visit_yul_block(node),
    yul.Break: lambda self, node: self.visit_yul_break(node),
    yul.Case: lambda self, node: self.visit_yul_case(node),
    yul.Continue: lambda self, node: self.visit_yul_continue(node),
    yul.ExpressionStatement: lambda self, node: self.visit_yul_expression_statement(
        node
    ),
    yul.ForLoop: lambda self, node: self.visit_yul_for_loop(node),
    yul.FunctionCall: lambda self, node: self.visit_yul_function_call(node),
    yul.FunctionDefinition: lambda self, node: self.visit_yul_function_definition(node),
    yul.Identifier: lambda self, node: self.visit_yul_identifier(node),
    yul.If: lambda self, node: self.visit_yul_if(node),
    yul.Leave: lambda self, node: self.visit_yul_leave(node),
    yul.Literal: lambda self, node: self.visit_yul_literal(node),
    yul.Switch: lambda self, node: self.visit_yul_switch(node),
    yul.TypedName: lambda self, node: self.visit_yul_typed_name(node),
    yul.VariableDeclaration: lambda self, node: self.visit_yul_variable_declaration(
        node
    ),
}


@dataclass(eq=True, frozen=True)
class DetectorResult:
    ir_node: IrAbc
    message: str
    related_info: Tuple[DetectorResult, ...] = field(default_factory=tuple)
    lsp_range: Optional[Tuple[int, int]] = field(default=None)


@dataclass
class DetectionResult:
    result: DetectorResult
    code: int
    string_id: str


@dataclass
class Detector:
    code: int
    string_id: str
    detector: Type[DetectorAbc]


class DetectorAbc(ABC):
    @abstractmethod
    def report(self) -> List[DetectorResult]:
        ...

    # declarations
    def visit_contract_definition(self, node: ContractDefinition):
        pass

    def visit_enum_definition(self, node: EnumDefinition):
        pass

    def visit_enum_value(self, node: EnumValue):
        pass

    def visit_error_definition(self, node: ErrorDefinition):
        pass

    def visit_event_definition(self, node: EventDefinition):
        pass

    def visit_function_definition(self, node: FunctionDefinition):
        pass

    def visit_modifier_definition(self, node: ModifierDefinition):
        pass

    def visit_struct_definition(self, node: StructDefinition):
        pass

    def visit_user_defined_value_type_definition(
        self, node: UserDefinedValueTypeDefinition
    ):
        pass

    def visit_variable_declaration(self, node: VariableDeclaration):
        pass

    # expressions
    def visit_assignment(self, node: Assignment):
        pass

    def visit_binary_operation(self, node: BinaryOperation):
        pass

    def visit_conditional(self, node: Conditional):
        pass

    def visit_elementary_type_name_expression(self, node: ElementaryTypeNameExpression):
        pass

    def visit_function_call(self, node: FunctionCall):
        pass

    def visit_function_call_options(self, node: FunctionCallOptions):
        pass

    def visit_identifier(self, node: Identifier):
        pass

    def visit_index_access(self, node: IndexAccess):
        pass

    def visit_index_range_access(self, node: IndexRangeAccess):
        pass

    def visit_literal(self, node: Literal):
        pass

    def visit_member_access(self, node: MemberAccess):
        pass

    def visit_new_expression(self, node: NewExpression):
        pass

    def visit_tuple_expression(self, node: TupleExpression):
        pass

    def visit_unary_operation(self, node: UnaryOperation):
        pass

    # meta
    def visit_identifier_path(self, node: IdentifierPath):
        pass

    def visit_import_directive(self, node: ImportDirective):
        pass

    def visit_inheritance_specifier(self, node: InheritanceSpecifier):
        pass

    def visit_modifier_invocation(self, node: ModifierInvocation):
        pass

    def visit_override_specifier(self, node: OverrideSpecifier):
        pass

    def visit_parameter_list(self, node: ParameterList):
        pass

    def visit_pragma_directive(self, node: PragmaDirective):
        pass

    def visit_source_unit(self, node: SourceUnit):
        pass

    def visit_structured_documentation(self, node: StructuredDocumentation):
        pass

    def visit_try_catch_clause(self, node: TryCatchClause):
        pass

    def visit_using_for_directive(self, node: UsingForDirective):
        pass

    # statements
    def visit_block(self, node: Block):
        pass

    def visit_break(self, node: Break):
        pass

    def visit_continue(self, node: Continue):
        pass

    def visit_do_while_statement(self, node: DoWhileStatement):
        pass

    def visit_emit_statement(self, node: EmitStatement):
        pass

    def visit_expression_statement(self, node: ExpressionStatement):
        pass

    def visit_for_statement(self, node: ForStatement):
        pass

    def visit_if_statement(self, node: IfStatement):
        pass

    def visit_inline_assembly(self, node: InlineAssembly):
        pass

    def visit_placeholder_statement(self, node: PlaceholderStatement):
        pass

    def visit_return(self, node: Return):
        pass

    def visit_revert_statement(self, node: RevertStatement):
        pass

    def visit_try_statement(self, node: TryStatement):
        pass

    def visit_unchecked_block(self, node: UncheckedBlock):
        pass

    def visit_variable_declaration_statement(self, node: VariableDeclarationStatement):
        pass

    def visit_while_statement(self, node: WhileStatement):
        pass

    # type names
    def visit_array_type_name(self, node: ArrayTypeName):
        pass

    def visit_elementary_type_name(self, node: ElementaryTypeName):
        pass

    def visit_function_type_name(self, node: FunctionTypeName):
        pass

    def visit_mapping(self, node: Mapping):
        pass

    def visit_user_defined_type_name(self, node: UserDefinedTypeName):
        pass

    # yul
    def visit_yul_assignment(self, node: yul.Assignment):
        pass

    def visit_yul_block(self, node: yul.Block):
        pass

    def visit_yul_break(self, node: yul.Break):
        pass

    def visit_yul_case(self, node: yul.Case):
        pass

    def visit_yul_continue(self, node: yul.Continue):
        pass

    def visit_yul_expression_statement(self, node: yul.ExpressionStatement):
        pass

    def visit_yul_for_loop(self, node: yul.ForLoop):
        pass

    def visit_yul_function_call(self, node: yul.FunctionCall):
        pass

    def visit_yul_function_definition(self, node: yul.FunctionDefinition):
        pass

    def visit_yul_identifier(self, node: yul.Identifier):
        pass

    def visit_yul_if(self, node: yul.If):
        pass

    def visit_yul_leave(self, node: yul.Leave):
        pass

    def visit_yul_literal(self, node: yul.Literal):
        pass

    def visit_yul_switch(self, node: yul.Switch):
        pass

    def visit_yul_typed_name(self, node: yul.TypedName):
        pass

    def visit_yul_variable_declaration(self, node: yul.VariableDeclaration):
        pass


def _get_enabled_detectors(config: WokeConfig) -> List[Detector]:
    ret = []
    for d in detectors:
        if (
            config.detectors.only is not None
            and d.string_id not in config.detectors.only
        ):
            continue
        if (
            d.string_id not in config.detectors.exclude
            and d.code not in config.detectors.exclude
        ):
            ret.append(d)
    return ret


def detect(
    config: WokeConfig,
    source_units: Dict[Path, SourceUnit],
    *,
    console: Optional[rich.console.Console] = None,
) -> List[DetectionResult]:
    def _detection_ignored(detection: DetectorResult) -> bool:
        return any(
            is_relative_to(detection.ir_node.file, p)
            for p in config.detectors.ignore_paths
        ) and all(_detection_ignored(d) for d in detection.related_info)

    results: DefaultDict[str, DefaultDict[str, List[DetectionResult]]] = defaultdict(
        lambda: defaultdict(list)
    )
    enabled_detectors = []
    enabled_detector_instances = []
    for d in _get_enabled_detectors(config):
        enabled_detectors.append(d)
        enabled_detector_instances.append(d.detector())

    path_to_source_unit_name: Dict[Path, str] = {}

    if console is not None:
        original_record = console.record
        console.record = False
        ctx_manager = console.status("[bold green]Running detectors...")
    else:
        original_record = True
        ctx_manager = nullcontext()
    with ctx_manager as s:
        for path, source_unit in source_units.items():
            if s is not None:
                s.update(f"[bold green]Detecting in {source_unit.source_unit_name}...")
            path_to_source_unit_name[path] = source_unit.source_unit_name
            for ir_node in source_unit:
                for d, detector_instance in zip(
                    enabled_detectors, enabled_detector_instances
                ):
                    visit_map[ir_node.__class__](detector_instance, ir_node)

    if console is not None:
        console.record = original_record

    for d, detector_instance in zip(enabled_detectors, enabled_detector_instances):
        for result in detector_instance.report():
            if not _detection_ignored(result):
                results[d.string_id][
                    path_to_source_unit_name[result.ir_node.file]
                ].append(
                    DetectionResult(
                        result=result,
                        code=d.code,
                        string_id=d.string_id,
                    )
                )

    ret = []
    sorted_detectors = sorted(results.keys())
    for detector_id in sorted_detectors:
        sorted_files = sorted(results[detector_id].keys())
        for file in sorted_files:
            ret.extend(
                sorted(
                    results[detector_id][file],
                    key=lambda d: d.result.ir_node.byte_location[0],
                )
            )
    return ret


def print_detectors(config: WokeConfig, theme: str = "monokai") -> None:
    applied_detectors = set()
    for d in _get_enabled_detectors(config):
        applied_detectors.add((d.string_id, d.detector.__doc__))

    detectors_list = "Using the following detectors:\n- " + "\n- ".join(
        f"{d[0]}\n\t{d[1]}" for d in sorted(applied_detectors)
    )
    woke.cli.console.console.print(
        Markdown(detectors_list, inline_code_theme=theme, inline_code_lexer="solidity")
    )


def print_detection(
    result: DetectionResult, theme: Union[str, SyntaxTheme] = "monokai"
) -> None:
    def print_result(
        detector_result: DetectorResult,
        tree: Optional[Tree],
        detector_id: Optional[str],
    ) -> Tree:
        source_unit = detector_result.ir_node
        while source_unit is not None:
            if isinstance(source_unit, SourceUnit):
                break
            source_unit = source_unit.parent
        assert isinstance(source_unit, SourceUnit)

        tmp_lines = re.split(b"(\r?\n)", source_unit.file_source)
        lines: List[bytes] = []
        for line in tmp_lines:
            if line in {b"\r\n", b"\n"}:
                lines[-1] += line
            else:
                lines.append(line)

        offset = 0
        line_index = 0
        while offset <= detector_result.ir_node.byte_location[0]:
            offset += len(lines[line_index])
            line_index += 1
        line_index -= 1

        source = ""
        start_line_index = max(0, line_index - 3)
        end_line_index = min(len(lines), line_index + 3)
        for i in range(start_line_index, end_line_index):
            source += lines[i].decode("utf-8")

        panel = Panel.fit(
            Syntax(
                source,
                "solidity",
                theme=theme,
                line_numbers=True,
                start_line=(start_line_index + 1),
                highlight_lines={line_index + 1},
            ),
            title=f"{detector_result.message} \[{detector_id}]"  # pyright: ignore reportInvalidStringEscapeSequence
            if detector_id is not None
            else detector_result.message,
            title_align="left",
            subtitle=source_unit.source_unit_name,
            subtitle_align="left",
        )

        if tree is None:
            t = Tree(panel)
        else:
            t = tree.add(panel)

        for additional_result in detector_result.related_info:
            print_result(additional_result, t, None)

        return t

    woke.cli.console.console.print("\n")
    tree = print_result(result.result, None, result.string_id)
    woke.cli.console.console.print(tree)


def detector(code: int, string_id: str):
    def decorator(d: Type[DetectorAbc]):
        detectors.append(Detector(code=code, string_id=string_id, detector=d))
        return d

    return decorator



from abc import ABC, abstractmethod
from typing import Any


class ProtocolAbc(ABC):
    @abstractmethod
    def __enter__(self):
        ...

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        ...

    @abstractmethod
    def send_recv(self, data: str) -> Any:
        ...



from functools import lru_cache, partial
from typing import Iterator, Optional, Set, Tuple

from woke.ast.enums import BinaryOpOperator, ModifiesStateFlag
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import AstNodeId, SolcBinaryOperation

from ..declaration.function_definition import FunctionDefinition
from ..reference_resolver import CallbackParams
from .abc import ExpressionAbc


class BinaryOperation(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcBinaryOperation
    _parent: SolidityAbc  # TODO: make this more specific

    _left_expression: ExpressionAbc
    _operator: BinaryOpOperator
    _right_expression: ExpressionAbc
    _function_id: Optional[AstNodeId]

    def __init__(
        self,
        init: IrInitTuple,
        binary_operation: SolcBinaryOperation,
        parent: SolidityAbc,
    ):
        super().__init__(init, binary_operation, parent)
        self._operator = binary_operation.operator
        self._left_expression = ExpressionAbc.from_ast(
            init, binary_operation.left_expression, self
        )
        self._right_expression = ExpressionAbc.from_ast(
            init, binary_operation.right_expression, self
        )
        self._function_id = binary_operation.function
        if self._function_id is not None:
            init.reference_resolver.register_post_process_callback(self._post_process)

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._left_expression
        yield from self._right_expression

    def _post_process(self, callback_params: CallbackParams):
        function = self.function
        assert function is not None
        function.register_reference(self)
        self._reference_resolver.register_destroy_callback(
            self.file, partial(self._destroy, function)
        )

    def _destroy(self, function: FunctionDefinition) -> None:
        function.unregister_reference(self)

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def operator(self) -> BinaryOpOperator:
        return self._operator

    @property
    def left_expression(self) -> ExpressionAbc:
        return self._left_expression

    @property
    def right_expression(self) -> ExpressionAbc:
        return self._right_expression

    @property
    def is_ref_to_state_variable(self) -> bool:
        return False

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return (
            self.left_expression.modifies_state | self.right_expression.modifies_state
        )

    @property
    def function(self) -> Optional[FunctionDefinition]:
        if self._function_id is None:
            return None
        node = self._reference_resolver.resolve_node(self._function_id, self._cu_hash)
        assert isinstance(node, FunctionDefinition)
        return node



import asyncio
import collections
import json
from typing import Optional, Union

from .protocol_structures import (
    NotificationMessage,
    RequestMessage,
    ResponseError,
    ResponseMessage,
)

# TODO Buffering messages


ENCODING = "utf-8"


class RpcProtocolError(Exception):
    pass


class RpcProtocol:
    """
    Json rpc communication
    """

    __port: int
    __reader: asyncio.StreamReader
    __writer: asyncio.StreamWriter
    __lock: asyncio.Lock

    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        self.__reader = reader
        self.__writer = writer
        self.__lock = asyncio.Lock()

    async def _read_message(self) -> dict:
        line = (await self.__reader.readline()).decode(ENCODING)
        # It is, read header then
        if line.startswith("Content-Length: ") and line.endswith("\r\n"):
            content_length = int(line.split(":")[-1])
        else:
            raise RpcProtocolError(f"Invalid HTTP header: {line}")
        # Skip unnecessary header part
        while line != b"\r\n":
            line = await self.__reader.readline()

        return json.loads(await self.__reader.readexactly(content_length))

    async def receive(
        self,
    ) -> Union[RequestMessage, NotificationMessage, ResponseMessage]:
        raw_message = await self._read_message()

        if "id" in raw_message:
            if "method" in raw_message:
                return RequestMessage.parse_obj(raw_message)
            else:
                return ResponseMessage.parse_obj(raw_message)
        return NotificationMessage.parse_obj(raw_message)

    async def _send(self, message: str) -> None:
        content_length = len(message)
        response = f"Content-Length: {content_length}\r\nContent-Type: application/vscode-jsonrpc; charset={ENCODING}\r\n\r\n{message}"
        async with self.__lock:
            self.__writer.write(response.encode(ENCODING))
            await self.__writer.drain()

    async def send(
        self,
        message: Union[
            ResponseMessage, RequestMessage, ResponseError, NotificationMessage
        ],
    ) -> None:
        await self._send(message.json(exclude_unset=True, by_alias=True))



from __future__ import annotations

import asyncio
import sys
import time
from pathlib import Path
from typing import TYPE_CHECKING, Set

import rich_click as click
from click.core import Context

if TYPE_CHECKING:
    from woke.config import WokeConfig


def write_config(config: WokeConfig) -> None:
    with config.project_root_path.joinpath("woke.toml").open("w") as f:
        f.write("[compiler.solc]\n")
        f.write('ignore_paths = ["node_modules", ".woke-build", "venv", "lib"]\n')
        f.write('include_paths = ["node_modules"]\n')
        if len(config.compiler.solc.remappings) > 0:
            f.write("remappings = [\n")
            for r in config.compiler.solc.remappings:
                f.write(f'    "{r}",\n')
            f.write("]\n")
        if config.compiler.solc.via_IR:
            f.write("via_IR = true\n")
        f.write("\n")

        if config.compiler.solc.optimizer.enabled:
            f.write("[compiler.solc.optimizer]\n")
            f.write("enabled = true\n")
            f.write(f"runs = {config.compiler.solc.optimizer.runs}\n")
            f.write("\n")

        f.write("[detectors]\n")
        f.write("exclude = []\n")
        f.write('ignore_paths = ["node_modules", ".woke-build", "venv", "lib"]\n')
        f.write("\n")

        f.write("[testing]\n")
        f.write('cmd = "anvil"\n')
        f.write("\n")

        f.write("[testing.anvil]\n")
        f.write(f'cmd_args = "{config.testing.anvil.cmd_args}"\n')
        f.write("\n")

        f.write("[testing.ganache]\n")
        f.write(f'cmd_args = "{config.testing.ganache.cmd_args}"\n')
        f.write("\n")

        f.write("[testing.hardhat]\n")
        f.write(f'cmd_args = "{config.testing.hardhat.cmd_args}"')


def update_gitignore(file: Path) -> None:
    if file.exists():
        lines = file.read_text().splitlines()
    else:
        lines = []

    new_lines = [
        ".woke-build",
        ".woke-logs",
        ".env",
        "pytypes",
        "__pycache__/",
        "*.py[cod]",
        ".hypothesis/",
        "woke-coverage.cov",
    ]

    new_lines = [l for l in new_lines if l not in lines]

    if len(new_lines) > 0:
        with file.open("a") as f:
            f.write("\n" + "\n".join(new_lines))


@click.group(name="init", invoke_without_command=True)
@click.option(
    "--force", "-f", is_flag=True, default=False, help="Force overwrite existing files."
)
@click.pass_context
def run_init(ctx: Context, force: bool):
    """Initialize project."""
    from woke.config import WokeConfig

    config = WokeConfig()
    config.load_configs()
    ctx.obj["config"] = config

    if ctx.invoked_subcommand is None:
        import subprocess

        from ..compiler import SolcOutputSelectionEnum, SolidityCompiler
        from ..compiler.solc_frontend import SolcOutputErrorSeverityEnum
        from ..development.pytypes_generator import TypeGenerator
        from ..utils.file_utils import copy_dir, is_relative_to
        from .console import console

        # create tests directory
        copy_dir(
            Path(__file__).parent.parent / "templates" / "tests",
            config.project_root_path / "tests",
            overwrite=force,
        )

        # create scripts directory
        copy_dir(
            Path(__file__).parent.parent / "templates" / "scripts",
            config.project_root_path / "scripts",
            overwrite=force,
        )

        # update .gitignore, --force is not needed
        update_gitignore(config.project_root_path / ".gitignore")

        # load foundry remappings, if foundry.toml exists
        if (config.project_root_path / "foundry.toml").exists():
            remappings = (
                subprocess.run(["forge", "remappings"], capture_output=True)
                .stdout.decode("utf-8")
                .splitlines()
            )
            config.update({"compiler": {"solc": {"remappings": remappings}}}, [])

        sol_files: Set[Path] = set()
        start = time.perf_counter()
        with console.status("[bold green]Searching for *.sol files...[/]"):
            for file in config.project_root_path.rglob("**/*.sol"):
                if (
                    not any(
                        is_relative_to(file, p)
                        for p in config.compiler.solc.ignore_paths
                    )
                    and file.is_file()
                ):
                    sol_files.add(file)
        end = time.perf_counter()
        console.log(
            f"[green]Found {len(sol_files)} *.sol files in [bold green]{end - start:.2f} s[/bold green][/]"
        )

        if len(sol_files) == 0:
            (config.project_root_path / "contracts").mkdir(exist_ok=True)
        else:
            compiler = SolidityCompiler(config)
            compiler.load(console=console)

            _, errors = asyncio.run(
                compiler.compile(
                    sol_files,
                    [SolcOutputSelectionEnum.ALL],
                    write_artifacts=True,
                    force_recompile=False,
                    console=console,
                    no_warnings=True,
                )
            )

            contract_size_limit = any(
                e
                for e in errors
                if e.severity == SolcOutputErrorSeverityEnum.WARNING
                and "Contract code size" in e.message
            )
            stack_too_deep = any(
                e
                for e in errors
                if e.severity == SolcOutputErrorSeverityEnum.ERROR
                and "Stack too deep" in e.message
            )

            if contract_size_limit or stack_too_deep:
                if stack_too_deep:
                    console.print(
                        "[yellow]Stack too deep error detected. Enabling optimizer.[/]"
                    )
                elif contract_size_limit:
                    console.print(
                        "[yellow]Contract size limit warning detected. Enabling optimizer.[/]"
                    )
                config.update(
                    {"compiler": {"solc": {"optimizer": {"enabled": True}}}}, []
                )

                _, errors = asyncio.run(
                    compiler.compile(
                        sol_files,
                        [SolcOutputSelectionEnum.ALL],
                        write_artifacts=True,
                        force_recompile=False,
                        console=console,
                        no_warnings=True,
                    )
                )
                stack_too_deep = any(
                    e
                    for e in errors
                    if e.severity == SolcOutputErrorSeverityEnum.ERROR
                    and "Stack too deep" in e.message
                )

                if stack_too_deep:
                    console.print(
                        "[yellow]Stack too deep error still detected. Enabling --via-ir.[/]"
                    )
                    config.update({"compiler": {"solc": {"via_IR": True}}}, [])

                    _, errors = asyncio.run(
                        compiler.compile(
                            sol_files,
                            [SolcOutputSelectionEnum.ALL],
                            write_artifacts=True,
                            force_recompile=False,
                            console=console,
                            no_warnings=True,
                        )
                    )
            start = time.perf_counter()
            with console.status("[bold green]Generating pytypes..."):
                type_generator = TypeGenerator(config, False)
                type_generator.generate_types(compiler)
            end = time.perf_counter()
            console.log(
                f"[green]Generated pytypes in [bold green]{end - start:.2f} s[/]"
            )

    if not (config.project_root_path / "woke.toml").exists() or force:
        write_config(config)


async def run_init_pytypes(
    config: WokeConfig, return_tx: bool, warnings: bool, watch: bool
):
    from watchdog.observers import Observer

    from ..compiler import SolcOutputSelectionEnum, SolidityCompiler
    from ..compiler.build_data_model import ProjectBuild, ProjectBuildInfo
    from ..compiler.compiler import CompilationFileSystemEventHandler
    from ..compiler.solc_frontend import SolcOutputErrorSeverityEnum
    from ..development.pytypes_generator import TypeGenerator
    from ..utils.file_utils import is_relative_to
    from .console import console

    def callback(build: ProjectBuild, build_info: ProjectBuildInfo):
        start = time.perf_counter()
        with console.status("[bold green]Generating pytypes..."):
            type_generator = TypeGenerator(config, return_tx)
            type_generator.generate_types(compiler)
        end = time.perf_counter()
        console.log(f"[green]Generated pytypes in [bold green]{end - start:.2f} s[/]")

    compiler = SolidityCompiler(config)

    sol_files: Set[Path] = set()
    start = time.perf_counter()
    with console.status("[bold green]Searching for *.sol files...[/]"):
        for file in config.project_root_path.rglob("**/*.sol"):
            if (
                not any(
                    is_relative_to(file, p) for p in config.compiler.solc.ignore_paths
                )
                and file.is_file()
            ):
                sol_files.add(file)
    end = time.perf_counter()
    console.log(
        f"[green]Found {len(sol_files)} *.sol files in [bold green]{end - start:.2f} s[/bold green][/]"
    )

    if watch:
        fs_handler = CompilationFileSystemEventHandler(
            config,
            sol_files,
            asyncio.get_event_loop(),
            compiler,
            [SolcOutputSelectionEnum.ALL],
            write_artifacts=True,
            console=console,
            no_warnings=not warnings,
        )
        fs_handler.register_callback(callback)

        observer = Observer()
        observer.schedule(
            fs_handler,
            str(config.project_root_path),
            recursive=True,
        )
        observer.start()
    else:
        fs_handler = None
        observer = None

    compiler.load(console=console)

    _, errors = await compiler.compile(
        sol_files,
        [SolcOutputSelectionEnum.ALL],
        write_artifacts=True,
        force_recompile=False,
        console=console,
        no_warnings=not warnings,
    )

    start = time.perf_counter()
    with console.status("[bold green]Generating pytypes..."):
        type_generator = TypeGenerator(config, return_tx)
        type_generator.generate_types(compiler)
    end = time.perf_counter()
    console.log(f"[green]Generated pytypes in [bold green]{end - start:.2f} s[/]")

    if watch:
        assert fs_handler is not None
        assert observer is not None
        try:
            await fs_handler.run()
        except KeyboardInterrupt:
            pass
        finally:
            observer.stop()
            observer.join()
    else:
        errored = any(e.severity == SolcOutputErrorSeverityEnum.ERROR for e in errors)
        if errored:
            sys.exit(1)


@run_init.command(name="pytypes")
@click.option(
    "--return-tx",
    is_flag=True,
    default=False,
    help="Return transaction objects from deploy functions instead of contract instances",
)
@click.option(
    "--warnings",
    "-W",
    is_flag=True,
    default=False,
    help="Print compilation warnings to console.",
)
@click.option(
    "--watch",
    "-w",
    is_flag=True,
    default=False,
    help="Watch for changes in the project and regenerate pytypes on change.",
)
@click.pass_context
def init_pytypes(ctx: Context, return_tx: bool, warnings: bool, watch: bool) -> None:
    """Generate Python types from Solidity sources."""
    config: WokeConfig = ctx.obj["config"]
    asyncio.run(run_init_pytypes(config, return_tx, warnings, watch))



from typing import List, Set

from woke.ast.enums import GlobalSymbolsEnum
from woke.ast.ir.expression.function_call import FunctionCall
from woke.ast.ir.expression.function_call_options import FunctionCallOptions
from woke.ast.ir.expression.member_access import MemberAccess

from .api import DetectorAbc, DetectorResult, detector


@detector(-1002, "function-call-options-not-called")
class FunctionCallOptionsNotCalledDetector(DetectorAbc):
    """
    Function with gas or value set actually is not called, e.g. `this.externalFunction.value(targetValue)` or `this.externalFunction{value: targetValue}`.
    """

    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def report(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_function_call_options(self, node: FunctionCallOptions):
        if not isinstance(node.parent, FunctionCall):
            self._detections.add(
                DetectorResult(node, "Function call options not called")
            )

    def visit_member_access(self, node: MemberAccess):
        if node.referenced_declaration in {
            GlobalSymbolsEnum.FUNCTION_GAS,
            GlobalSymbolsEnum.FUNCTION_VALUE,
        }:
            parent = node.parent
            assert isinstance(parent, FunctionCall)

            if not isinstance(parent.parent, FunctionCall):
                if node.referenced_declaration == GlobalSymbolsEnum.FUNCTION_GAS:
                    self._detections.add(
                        DetectorResult(node, "Function with gas not called")
                    )
                else:
                    self._detections.add(
                        DetectorResult(node, "Function with value not called")
                    )



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Tuple, Union

from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import YulFunctionCall, YulIdentifier, YulLiteral

from .abc import YulAbc
from .identifier import Identifier
from .literal import Literal

if TYPE_CHECKING:
    from .assignment import Assignment
    from .expression_statement import ExpressionStatement
    from .for_loop import ForLoop
    from .if_statement import If
    from .switch import Switch
    from .variable_declaration import VariableDeclaration


class FunctionCall(YulAbc):
    """
    TBD
    """

    _parent: Union[
        Assignment,
        ExpressionStatement,
        ForLoop,
        If,
        Switch,
        VariableDeclaration,
        FunctionCall,
    ]
    _arguments: List[Union["FunctionCall", Identifier, Literal]]
    _function_name: Identifier

    def __init__(
        self, init: IrInitTuple, function_call: YulFunctionCall, parent: YulAbc
    ):
        super().__init__(init, function_call, parent)
        self._function_name = Identifier(init, function_call.function_name, self)
        self._arguments = []
        for argument in function_call.arguments:
            if isinstance(argument, YulFunctionCall):
                self._arguments.append(FunctionCall(init, argument, self))
            elif isinstance(argument, YulIdentifier):
                self._arguments.append(Identifier(init, argument, self))
            elif isinstance(argument, YulLiteral):
                self._arguments.append(Literal(init, argument, self))
            else:
                assert False, f"Unexpected type: {type(argument)}"

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        yield from self._function_name
        for argument in self._arguments:
            yield from argument

    @property
    def parent(
        self,
    ) -> Union[
        Assignment,
        ExpressionStatement,
        ForLoop,
        If,
        Switch,
        VariableDeclaration,
        FunctionCall,
    ]:
        return self._parent

    @property
    def arguments(self) -> Tuple[Union["FunctionCall", Identifier, Literal], ...]:
        return tuple(self._arguments)

    @property
    def function_name(self) -> Identifier:
        return self._function_name



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, Union

from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import (
    YulExpressionStatement,
    YulFunctionCall,
    YulIdentifier,
    YulLiteral,
)

from .abc import YulAbc, YulStatementAbc
from .function_call import FunctionCall
from .identifier import Identifier
from .literal import Literal

if TYPE_CHECKING:
    from .block import Block


class ExpressionStatement(YulStatementAbc):
    """
    TBD
    """

    _parent: Block
    _expression: Union[FunctionCall, Identifier, Literal]

    def __init__(
        self,
        init: IrInitTuple,
        expression_statement: YulExpressionStatement,
        parent: YulAbc,
    ):
        super().__init__(init, expression_statement, parent)
        if isinstance(expression_statement.expression, YulFunctionCall):
            self._expression = FunctionCall(init, expression_statement.expression, self)
        elif isinstance(expression_statement.expression, YulIdentifier):
            self._expression = Identifier(init, expression_statement.expression, self)
        elif isinstance(expression_statement.expression, YulLiteral):
            self._expression = Literal(init, expression_statement.expression, self)
        else:
            assert False, f"Unexpected type: {type(expression_statement.expression)}"

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        yield from self._expression

    @property
    def parent(self) -> Block:
        return self._parent

    @property
    def expression(self) -> Union[FunctionCall, Identifier, Literal]:
        return self._expression



import logging
from pathlib import Path
from typing import List, Optional, Tuple, Union

import woke.ast.ir.yul as yul

from ...ast.enums import GlobalSymbolsEnum
from ...ast.ir.abc import IrAbc
from ...ast.ir.declaration.abc import DeclarationAbc
from ...ast.ir.declaration.contract_definition import ContractDefinition
from ...ast.ir.expression.binary_operation import BinaryOperation
from ...ast.ir.expression.identifier import Identifier
from ...ast.ir.expression.member_access import MemberAccess
from ...ast.ir.expression.unary_operation import UnaryOperation
from ...ast.ir.meta.identifier_path import IdentifierPath
from ...ast.ir.type_name.user_defined_type_name import UserDefinedTypeName
from ...core.solidity_version import SemanticVersion
from ..common_structures import (
    MarkupContent,
    MarkupKind,
    MessageType,
    Position,
    Range,
    TextDocumentPositionParams,
    TextDocumentRegistrationOptions,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from ..context import LspContext
from ..lsp_data_model import LspModel
from ..utils import position_within_range, uri_to_path
from ..utils.position import changes_to_byte_offset

logger = logging.getLogger(__name__)


class HoverClientCapabilities(LspModel):
    dynamic_registration: Optional[bool]
    content_format: Optional[List[MarkupKind]]


class HoverOptions(WorkDoneProgressOptions):
    pass


class HoverRegistrationOptions(TextDocumentRegistrationOptions, HoverOptions):
    pass


class HoverParams(TextDocumentPositionParams, WorkDoneProgressParams):
    pass


class MarkedString(LspModel):
    language: str
    value: str


MarkedStringType = Union[str, MarkedString]


class Hover(LspModel):
    contents: Union[MarkedStringType, List[MarkedStringType], MarkupContent]
    range: Optional[Range]


def _append_openzeppelin_docs(
    node: ContractDefinition, version: SemanticVersion
) -> str:
    source_unit = node.parent
    version_string = f"{version.major}.x"
    url_base = f"https://docs.openzeppelin.com/contracts/{version_string}"
    doc_url = None
    api_doc_url = None

    if "openzeppelin/contracts/access" in source_unit.source_unit_name:
        doc_url = "access-control"
        api_doc_url = "access"
    elif "openzeppelin/contracts/crosschain" in source_unit.source_unit_name:
        doc_url = "crosschain"
        api_doc_url = "crosschain"
    elif "openzeppelin/contracts/finance" in source_unit.source_unit_name:
        if node.name == "PaymentSplitter":
            doc_url = "utilities#payment"
        api_doc_url = "finance"
    elif "openzeppelin/contracts/governance" in source_unit.source_unit_name:
        doc_url = "governance"
        api_doc_url = "governance"
    elif (
        "openzeppelin/contracts/interfaces" in source_unit.source_unit_name
        or "openzeppelin/contracts/token" in source_unit.source_unit_name
    ):
        if "ERC20" in node.name:
            doc_url = "erc20"
            api_doc_url = "token/erc20"
        elif "ERC721" in node.name:
            doc_url = "erc721"
            api_doc_url = "token/erc721"
        elif "ERC777" in node.name:
            doc_url = "erc777"
            api_doc_url = "token/erc777"
        elif "ERC1155" in node.name:
            doc_url = "erc1155"
            api_doc_url = "token/erc1155"
        elif node.name.startswith("I"):
            api_doc_url = "interfaces"
        else:
            api_doc_url = "token/common"
    elif "openzeppelin/contracts/metatx" in source_unit.source_unit_name:
        api_doc_url = "metatx"
    elif "openzeppelin/contracts/proxy" in source_unit.source_unit_name:
        api_doc_url = "proxy"
    elif "openzeppelin/contracts/security" in source_unit.source_unit_name:
        if node.name == "PullPayment":
            doc_url = "utilities#payment"
        api_doc_url = "security"
    elif "openzeppelin/contracts/utils" in source_unit.source_unit_name:
        doc_url = "utilities"
        api_doc_url = "utils"
    elif "openzeppelin/contracts/GSN" in source_unit.source_unit_name:
        doc_url = "gsn"
        api_doc_url = "gsn"
    elif "openzeppelin/contracts/cryptography" in source_unit.source_unit_name:
        doc_url = "utilities#cryptography"
        api_doc_url = "cryptography"
    elif "openzeppelin/contracts/drafts" in source_unit.source_unit_name:
        api_doc_url = "drafts"
    elif "openzeppelin/contracts/math" in source_unit.source_unit_name:
        doc_url = "utilities#math"
        api_doc_url = "math"
    elif "openzeppelin/contracts/payment" in source_unit.source_unit_name:
        doc_url = "utilities#payment"
        api_doc_url = "payment"
    elif "openzeppelin/contracts/presets" in source_unit.source_unit_name:
        if "ERC20" in node.name:
            doc_url = "erc20#Presets"
        elif "ERC721" in node.name:
            doc_url = "erc721#Presets"
        elif "ERC777" in node.name:
            doc_url = "erc777#Presets"
        elif "ERC1155" in node.name:
            doc_url = "erc1155#Presets"
        api_doc_url = "presets"

    ret = ""
    if doc_url is not None:
        ret += f"\n\n[OpenZeppelin documentation]({url_base}/{doc_url})"
    if api_doc_url is not None:
        ret += f"\n\n[OpenZeppelin API documentation]({url_base}/api/{api_doc_url}#{node.name})"
    return ret


yul_definitions = {
    "stop": "// stop execution, identical to return(0, 0)\nstop()",
    "add": "// x + y\nadd(x, y)",
    "sub": "// x - y\nsub(x, y)",
    "mul": "// x * y\nmul(x, y)",
    "div": "// x / y, 0 if y == 0\ndiv(x, y)",
    "sdiv": "// signed x / y, 0 if y == 0\nsdiv(x, y)",
    "mod": "// x % y, 0 if y == 0\nmod(x, y)",
    "smod": "// signed x % y, 0 if y == 0\nsmod(x, y)",
    "exp": "// x ** y\nexp(x, y)",
    "not": "// bitwise NOT\nnot(x)",
    "lt": "// 1 if x < y, 0 otherwise\nlt(x, y)",
    "gt": "// 1 if x > y, 0 otherwise\ngt(x, y)",
    "slt": "// 1 if signed x < signed y, 0 otherwise\nslt(x, y)",
    "sgt": "// 1 if signed x > signed y, 0 otherwise\nsgt(x, y)",
    "eq": "// 1 if x == y, 0 otherwise\neq(x, y)",
    "iszero": "// 1 if x == 0, 0 otherwise\niszero(x)",
    "and": "// bitwise AND\nand(x, y)",
    "or": "// bitwise OR\nor(x, y)",
    "xor": "// bitwise XOR\nxor(x, y)",
    "byte": "// retrieve the nth most significant byte of x\nbyte(n, x)",
    "shl": "// x << y\nshl(x, y)",
    "shr": "// x >> y\nshr(x, y)",
    "sar": "// signed x >> y\nsar(x, y)",
    "addmod": "// (x + y) % m, 0 if m == 0\naddmod(x, y, m)",
    "mulmod": "// (x * y) % m, 0 if m == 0\nmulmod(x, y, m)",
    "signextend": "// sign extend from (i * 8 + 7)th bit counting from least significant\nsignextend(i, x)",
    "keccak256": "// compute the Keccak-256 hash\nkeccak256(memStart, length)",
    "pc": "// current program counter\npc()",
    "pop": "// discard value returned by another instruction\npop(x)",
    "mload": "// load word from memory\nmload(memStart)",
    "mstore": "// store word to memory\nmstore(memStart, x)",
    "mstore8": "// store single byte to memory\nmstore8(memStart, x)",
    "sload": "// load word from storage\nsload(storageSlot)",
    "sstore": "// store word to storage\nsstore(storageSlot, x)",
    "msize": "// size of memory\nmsize()",
    "gas": "// remaining gas\ngas()",
    "address": "// address of the current contract / execution context\naddress()",
    "balance": "// wei balance of an address\nbalance(address)",
    "selfbalance": "// wei balance of the current contract / execution context, same as balance(address()) but cheaper\nselfbalance()",
    "caller": "// caller (excluding delegatecall), same as `msg.sender` in Solidity\ncaller()",
    "callvalue": "// wei sent with the call, same as `msg.value` in Solidity\ncallvalue()",
    "calldataload": "// load word from call data\ncalldataload(callDataStart)",
    "calldatasize": "// size of call data in bytes\ncalldatasize()",
    "calldatacopy": "// copy call data to memory\ncalldatacopy(memStart, callDataStart, length)",
    "codesize": "// size of code of the current contract / execution context\ncodesize()",
    "codecopy": "// copy code to memory\ncodecopy(memStart, codeStart, length)",
    "extcodesize": "// size of code at address\nextcodesize(address)",
    "extcodecopy": "// copy code at address to memory\nextcodecopy(address, memStart, codeStart, length)",
    "returndatasize": "// size of return data buffer\nreturndatasize()",
    "returndatacopy": "// copy return data to memory\nreturndatacopy(memStart, returnDataStart, length)",
    "extcodehash": "// Keccak-256 hash of code at address\nextcodehash(address)",
    "create": "// create contract with code from memory and value in wei, return new address or 0 on error\ncreate(value, memStart, length)",
    "create2": "// create contract from salt with code from memory and value in wei, return new address or 0 on error\ncreate2(value, memStart, length, salt)",
    "call": "// call contract at address providing gas, value in wei, call data from memory, storing return data to memory, return 1 on success or 0 on error\ncall(gas, address, value, memStartIn, memLengthIn, memStartOut, memLengthOut)",
    "callcode": "// call contract at address preserving execution context, providing gas, value in wei, call data from memory, storing return data to memory, return 1 on success or 0 on error\ncallcode(gas, address, value, memStartIn, memLengthIn, memStartOut, memLengthOut)",
    "delegatecall": "// call contract at address preserving execution context, current caller and call value, providing gas, call data from memory, storing return data to memory, return 1 on success or 0 on error\ndelegatecall(gas, address, memStartIn, memLengthIn, memStartOut, memLengthOut)",
    "staticcall": "// call contract at address disallowing state modifications, providing gas, call data from memory, storing return data to memory, return 1 on success or 0 on error\nstaticcall(gas, address, memStartIn, memLengthIn, memStartOut, memLengthOut)",
    "return": "// end execution, copy memory to return data buffer\nreturn(memStart, length)",
    "revert": "// end execution, revert state changes, copy memory to return data buffer\nrevert(memStart, length)",
    "selfdestruct": "// end execution, mark current contract to be destroyed, sending funds to address\nselfdestruct(address)",
    "invalid": "// end execution with invalid instruction, consume all remaining gas\ninvalid()",
    "log0": "// log without topics data from memory\nlog0(memStart, length)",
    "log1": "// log with one topic data from memory\nlog1(memStart, length, topic1)",
    "log2": "// log with two topics data from memory\nlog2(memStart, length, topic1, topic2)",
    "log3": "// log with three topics data from memory\nlog3(memStart, length, topic1, topic2, topic3)",
    "log4": "// log with four topics data from memory\nlog4(memStart, length, topic1, topic2, topic3, topic4)",
    "chainid": "// current chain ID\nchainid()",
    "basefee": "// current block's base fee\nbasefee()",
    "origin": "// sender of the transaction, same as `tx.origin` in Solidity\norigin()",
    "gasprice": "// gas price of the transaction, same as `tx.gasprice` in Solidity\ngasprice()",
    "blockhash": "// hash of the given block number, only for last 256 blocks excluding current\nblockhash(blockNumber)",
    "coinbase": "// current block's miner address\ncoinbase()",
    "timestamp": "// timestamp of the current block in seconds since the epoch\ntimestamp()",
    "number": "// current block number\nnumber()",
    "difficulty": "// current block's difficulty\ndifficulty()",
    "gaslimit": "// current block's gas limit\ngaslimit()",
    "prevrandao": "// previous block's RANDAO value\nprevrandao()",
}


def _get_results_from_node(
    original_node: IrAbc,
    position: Position,
    byte_offset: int,
    context: LspContext,
    node_name_location: Optional[Tuple[int, int]],
) -> Optional[Tuple[str, Tuple[int, int]]]:
    original_node_location: Tuple[int, int] = original_node.byte_location
    logger.debug(f"Found node {original_node}")

    if isinstance(original_node, DeclarationAbc):
        assert node_name_location is not None
        name_location_range = context.compiler.get_range_from_byte_offsets(
            original_node.file, node_name_location
        )
        if not position_within_range(position, name_location_range):
            return None

    if isinstance(original_node, (Identifier, MemberAccess)):
        node = original_node.referenced_declaration
        if node is None:
            return None
    elif isinstance(original_node, (IdentifierPath, UserDefinedTypeName)):
        part = original_node.identifier_path_part_at(byte_offset)
        if part is None:
            return None
        node = part.referenced_declaration
        original_node_location = part.byte_location
    elif isinstance(original_node, yul.Identifier):
        external_reference = original_node.external_reference
        if external_reference is not None:
            node = external_reference.referenced_declaration
            original_node_location = external_reference.byte_location
        else:
            node = original_node
    elif (
        isinstance(original_node, (UnaryOperation, BinaryOperation))
        and original_node.function is not None
    ):
        node = original_node.function
    else:
        node = original_node

    if isinstance(node, DeclarationAbc):
        value = "```solidity\n" + node.declaration_string + "\n```"
        if (
            isinstance(node, ContractDefinition)
            and context.openzeppelin_contracts_version is not None
            and context.openzeppelin_contracts_version >= "2.0.0"
        ):
            value += _append_openzeppelin_docs(
                node, context.openzeppelin_contracts_version
            )

        return value, original_node_location
    elif isinstance(node, yul.Identifier):
        if node.name in yul_definitions:
            return (
                f"```solidity\n{yul_definitions[node.name]}\n```",
                original_node_location,
            )
    return None


async def _get_hover_from_cache(path: Path, position: Position, context: LspContext):
    new_byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, position.line, position.character
    )
    backward_changes = context.compiler.get_last_compilation_backward_changes(path)
    if backward_changes is None:
        raise Exception("No backward changes found")
    changes_before = backward_changes[0:new_byte_offset]
    old_byte_offset = changes_to_byte_offset(changes_before) + new_byte_offset

    tree = context.compiler.last_compilation_interval_trees[path]

    intervals = tree.at(old_byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    if len(nodes) == 0:
        raise ValueError(f"Could not find node at {old_byte_offset}")

    node = max(nodes, key=lambda n: n.ast_tree_depth)

    if isinstance(node, DeclarationAbc):
        location = node.name_location
        forward_changes = context.compiler.get_last_compilation_forward_changes(path)
        if forward_changes is None:
            raise Exception("No forward changes found")
        new_start = (
            changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
        )
        new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
        node_name_location = (new_start, new_end)
    else:
        node_name_location = None

    result = _get_results_from_node(
        node, position, old_byte_offset, context, node_name_location
    )
    if result is None:
        return None

    value, location = result
    forward_changes = context.compiler.get_last_compilation_forward_changes(path)
    if forward_changes is None:
        raise Exception("No forward changes found")
    new_start = changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
    new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
    return Hover(
        contents=MarkupContent(
            kind=MarkupKind.MARKDOWN,
            value=value,
        ),
        range=context.compiler.get_range_from_byte_offsets(path, (new_start, new_end)),
    )


async def hover(context: LspContext, params: HoverParams) -> Optional[Hover]:
    logger.debug(
        f"Hover for file {params.text_document.uri} at position {params.position} requested"
    )

    path = uri_to_path(params.text_document.uri).resolve()
    if (
        path not in context.compiler.interval_trees
        or not context.compiler.output_ready.is_set()
    ):
        try:
            return await _get_hover_from_cache(path, params.position, context)
        except Exception:
            pass

    await context.compiler.output_ready.wait()
    if path not in context.compiler.interval_trees:
        return None

    tree = context.compiler.interval_trees[path]

    byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, params.position.line, params.position.character
    )
    intervals = tree.at(byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    logger.debug(f"Found {len(nodes)} nodes at position {params.position}")
    if len(nodes) == 0:
        return None

    node = max(nodes, key=lambda n: n.ast_tree_depth)

    if isinstance(node, DeclarationAbc):
        node_name_location = node.name_location
    else:
        node_name_location = None

    result = _get_results_from_node(
        node, params.position, byte_offset, context, node_name_location
    )
    if result is None:
        return None

    value, original_node_location = result
    return Hover(
        contents=MarkupContent(
            kind=MarkupKind.MARKDOWN,
            value=value,
        ),
        range=context.compiler.get_range_from_byte_offsets(
            path, original_node_location
        ),
    )



from pathlib import Path, PurePath, PurePosixPath
from typing import List

from woke.config import WokeConfig
from woke.config.data_model import SolcRemapping


class SourceUnitNameResolver:
    """
    The main purpose of this class is to convert an import string into a source unit name. This is done with respect
    to https://docs.soliditylang.org/en/v0.8.11/path-resolution.html.
    """

    __config: WokeConfig

    def __init__(self, woke_config: WokeConfig):
        self.__config = woke_config

    def apply_remapping(self, parent_source_unit: str, source_unit_name: str) -> str:
        """
        Try to apply a remapping and return a source unit name. Up to one remapping can be applied to a single import.
        It is the longest one. In case of multiple remappings with the same length, the one specified last wins.
        """
        matching_remappings: List[SolcRemapping] = []
        for remapping in self.__config.compiler.solc.remappings:
            context, prefix, target = remapping
            context_matches = (context is None) or (
                context is not None and parent_source_unit.startswith(context)
            )
            if context_matches and source_unit_name.startswith(prefix):
                matching_remappings.append(remapping)

        if len(matching_remappings) == 0:
            return source_unit_name

        matching_remappings.sort(key=lambda r: len(r.context or ""), reverse=True)

        # choose the remapping with the longest context
        # if there are multiple remappings with the same context, choose the last one
        l = len(matching_remappings[0].prefix)
        target_remapping = matching_remappings[-1]
        for i in range(1, len(matching_remappings)):
            if len(matching_remappings[i].prefix) != l:
                target_remapping = matching_remappings[i - 1]
                break

        return source_unit_name.replace(
            str(target_remapping.prefix), target_remapping.target or "", 1
        )

    def __resolve_direct_import(self, parent_source_unit: str, import_str: str) -> str:
        """
        Return a source unit name of a direct import in the file with given source unit name.
        """
        return self.apply_remapping(parent_source_unit, import_str)

    def __resolve_relative_import(
        self, parent_source_unit: str, import_str: str
    ) -> str:
        """
        Return a source unit name of a relative import in the file with given source unit name.
        """
        import_parts = [part for part in import_str.split("/") if part not in {"", "."}]
        parent_parts = parent_source_unit.split("/")

        while len(parent_parts) > 0 and parent_parts[-1] == "":
            parent_parts.pop()
        if len(parent_parts) > 0:
            parent_parts.pop()
        while len(parent_parts) > 0 and parent_parts[-1] == "":
            parent_parts.pop()

        normalized_import_parts = []
        for part in import_parts:
            if part == "..":
                if (
                    len(normalized_import_parts) == 0
                    or normalized_import_parts[-1] == ".."
                ):
                    normalized_import_parts.append("..")
                else:
                    normalized_import_parts.pop()
            else:
                normalized_import_parts.append(part)

        for no, part in enumerate(normalized_import_parts):
            if part == "..":
                while len(parent_parts) > 0 and parent_parts[-1] == "":
                    parent_parts.pop()
                if len(parent_parts) > 0:
                    parent_parts.pop()
            else:
                normalized_import_parts = normalized_import_parts[no:]
                break

        if len(parent_parts) > 0:
            source_unit_name = (
                "/".join(parent_parts) + "/" + "/".join(normalized_import_parts)
            )
        else:
            source_unit_name = "/".join(normalized_import_parts)
        return self.apply_remapping(parent_source_unit, source_unit_name)

    def resolve_import(self, parent_source_unit: str, import_str: str) -> str:
        """
        Resolve a source unit name of an import in the file with given source unit name.
        """
        if import_str.startswith((".", "..")):
            return self.__resolve_relative_import(parent_source_unit, import_str)
        return self.__resolve_direct_import(parent_source_unit, import_str)

    def resolve_cmdline_arg(self, arg: str) -> str:
        """
        Return a source unit name of the file provided as a command-line argument.
        """
        path = Path(arg).resolve()
        pure_path = PurePath(path)
        return str(
            PurePosixPath(pure_path.relative_to(self.__config.project_root_path))
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from ...types import (
    Address,
    Bool,
    Bytes,
    Fixed,
    FixedBytes,
    Int,
    String,
    Type,
    UFixed,
    UInt,
)

if TYPE_CHECKING:
    from ..declaration.user_defined_value_type_definition import (
        UserDefinedValueTypeDefinition,
    )
    from ..declaration.variable_declaration import VariableDeclaration
    from ..expression.elementary_type_name_expression import (
        ElementaryTypeNameExpression,
    )
    from ..expression.new_expression import NewExpression
    from ..meta.using_for_directive import UsingForDirective
    from .array_type_name import ArrayTypeName
    from .mapping import Mapping

from woke.ast.enums import StateMutability
from woke.ast.ir.abc import IrAbc, SolidityAbc
from woke.ast.ir.type_name.abc import TypeNameAbc
from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import SolcElementaryTypeName


class ElementaryTypeName(TypeNameAbc):
    """
    Elementary type name.

    !!! example
        An elementary type name can be used:

        - inside a [VariableDeclaration][woke.ast.ir.declaration.variable_declaration.VariableDeclaration]:
            - both occurrences of `:::solidity uint` in line 1,
            - `:::solidity int` in line 1,
            - `:::solidity string` in line 10,
            - the first occurrence of `:::solidity bytes` in line 15,
        - inside a [UserDefinedValueTypeDefinition][woke.ast.ir.declaration.user_defined_value_type_definition.UserDefinedValueTypeDefinition]:
            - `:::solidity int` in line 7,
        - inside an [ElementaryTypeNameExpression][woke.ast.ir.expression.elementary_type_name_expression.ElementaryTypeNameExpression]:
            - `:::solidity int` in line 2,
        - inside a [NewExpression][woke.ast.ir.expression.new_expression.NewExpression]:
            - the second occurrence of `:::solidity bytes` in line 15,
        - inside a [UsingForDirective][woke.ast.ir.meta.using_for_directive.UsingForDirective]:
            - `:::solidity uint` in line 5,
        - inside an [ArrayTypeName][woke.ast.ir.type_name.array_type_name.ArrayTypeName]:
            - `:::solidity uint` in line 11,
        - inside a [Mapping][woke.ast.ir.type_name.mapping.Mapping]:
            - `:::solidity address` in line 12.

        ```solidity linenums="1"
        function add(uint a, uint b) pure returns(int) {
            return int(a + b);
        }

        using {add} for uint;

        type MyInt is int;

        contract C {
            string public str;
            uint[10] arr;
            mapping(address => MyInt) map;

            function foo() public pure {
                bytes memory b = new bytes(10);
            }
        }
        ```
    """

    _ast_node: SolcElementaryTypeName
    _parent: Union[
        VariableDeclaration,
        UserDefinedValueTypeDefinition,
        ElementaryTypeNameExpression,
        NewExpression,
        UsingForDirective,
        ArrayTypeName,
        Mapping,
    ]

    _name: str
    _state_mutability: Optional[StateMutability]

    def __init__(
        self,
        init: IrInitTuple,
        elementary_type_name: SolcElementaryTypeName,
        parent: SolidityAbc,
    ):
        super().__init__(init, elementary_type_name, parent)
        self._name = elementary_type_name.name
        self._state_mutability = elementary_type_name.state_mutability

        from woke.ast.ir.expression.elementary_type_name_expression import (
            ElementaryTypeNameExpression,
        )

        # fix missing type descriptions in AST
        if self._type_descriptions.type_identifier is None and isinstance(
            parent, ElementaryTypeNameExpression
        ):
            self._type_descriptions = parent._type_descriptions

    @property
    def parent(
        self,
    ) -> Union[
        VariableDeclaration,
        UserDefinedValueTypeDefinition,
        ElementaryTypeNameExpression,
        NewExpression,
        UsingForDirective,
        ArrayTypeName,
        Mapping,
    ]:
        """
        When the parent is a [NewExpression][woke.ast.ir.expression.new_expression.NewExpression], this can only be `bytes` or `string`.
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def type(
        self,
    ) -> Union[
        Address, Bool, Int, UInt, Fixed, UFixed, String, Bytes, FixedBytes, Type
    ]:
        """
        Returns either the generic [Type][woke.ast.types.Type] expression type (this is the case of a type conversion, for example `:::solidity address(0)`) or directly one of the elementary expression types.
        Returns:
            Type description.
        """
        t = super().type
        if not isinstance(
            t,
            (Address, Bool, Int, UInt, Fixed, UFixed, String, Bytes, FixedBytes, Type),
        ):
            raise TypeError(f"Unexpected type {t} {self.source}")
        assert isinstance(
            t,
            (Address, Bool, Int, UInt, Fixed, UFixed, String, Bytes, FixedBytes, Type),
        )
        return t

    @property
    def name(self) -> str:
        """
        !!! example
            For example `uint256`, `bool`, `string`, `bytes1` or `address`.

        !!! tip
            Instead of working with the name, it may be better to use the [type][woke.ast.ir.type_name.elementary_type_name.ElementaryTypeName.type] property.
        Returns:
            Name of the elementary type.
        """
        return self._name

    @property
    def state_mutability(self) -> Optional[StateMutability]:
        """
        Is only set for `address` as either [StateMutability.PAYABLE][woke.ast.enums.StateMutability.PAYABLE] or [StateMutability.NONPAYABLE][woke.ast.enums.StateMutability.NONPAYABLE].
        Returns:
            State mutability of the `address` type.
        """
        return self._state_mutability



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Tuple, Union

from woke.ast.ir.utils import IrInitTuple
from woke.ast.nodes import (
    YulAssignment,
    YulBlock,
    YulBreak,
    YulContinue,
    YulExpressionStatement,
    YulForLoop,
    YulFunctionDefinition,
    YulIf,
    YulLeave,
    YulSwitch,
    YulVariableDeclaration,
)

from .abc import YulAbc, YulStatementAbc

if TYPE_CHECKING:
    from woke.ast.ir.statement.inline_assembly import InlineAssembly

    from .assignment import Assignment
    from .break_statement import Break
    from .case_statement import Case
    from .continue_statement import Continue
    from .expression_statement import ExpressionStatement
    from .for_loop import ForLoop
    from .function_definition import FunctionDefinition
    from .if_statement import If
    from .leave import Leave
    from .switch import Switch
    from .variable_declaration import VariableDeclaration


class Block(YulStatementAbc):
    """
    TBD
    """

    _parent: Union[InlineAssembly, Block, ForLoop, FunctionDefinition, If, Case]
    _statements: List[
        Union[
            Assignment,
            "Block",
            Break,
            Continue,
            ExpressionStatement,
            Leave,
            ForLoop,
            FunctionDefinition,
            If,
            Switch,
            VariableDeclaration,
        ]
    ]

    def __init__(
        self, init: IrInitTuple, block: YulBlock, parent: Union[InlineAssembly, YulAbc]
    ):
        from .assignment import Assignment
        from .break_statement import Break
        from .continue_statement import Continue
        from .expression_statement import ExpressionStatement
        from .for_loop import ForLoop
        from .function_definition import FunctionDefinition
        from .if_statement import If
        from .leave import Leave
        from .switch import Switch
        from .variable_declaration import VariableDeclaration

        super().__init__(init, block, parent)
        self._statements = []
        for statement in block.statements:
            if isinstance(statement, YulAssignment):
                self._statements.append(Assignment(init, statement, self))
            elif isinstance(statement, YulBlock):
                self._statements.append(Block(init, statement, self))
            elif isinstance(statement, YulBreak):
                self._statements.append(Break(init, statement, self))
            elif isinstance(statement, YulContinue):
                self._statements.append(Continue(init, statement, self))
            elif isinstance(statement, YulExpressionStatement):
                self._statements.append(ExpressionStatement(init, statement, self))
            elif isinstance(statement, YulLeave):
                self._statements.append(Leave(init, statement, self))
            elif isinstance(statement, YulForLoop):
                self._statements.append(ForLoop(init, statement, self))
            elif isinstance(statement, YulFunctionDefinition):
                self._statements.append(FunctionDefinition(init, statement, self))
            elif isinstance(statement, YulIf):
                self._statements.append(If(init, statement, self))
            elif isinstance(statement, YulSwitch):
                self._statements.append(Switch(init, statement, self))
            elif isinstance(statement, YulVariableDeclaration):
                self._statements.append(VariableDeclaration(init, statement, self))
            else:
                assert False, f"Unexpected type: {type(statement)}"

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        for statement in self._statements:
            yield from statement

    @property
    def parent(
        self,
    ) -> Union[InlineAssembly, Block, ForLoop, FunctionDefinition, If, Case]:
        return self._parent

    @property
    def statements(
        self,
    ) -> Tuple[
        Union[
            Assignment,
            "Block",
            Break,
            Continue,
            ExpressionStatement,
            Leave,
            ForLoop,
            FunctionDefinition,
            If,
            Switch,
            VariableDeclaration,
        ],
        ...,
    ]:
        return tuple(self._statements)



import multiprocessing
import sys
from pathlib import Path
from typing import Callable, Iterable, Tuple

import rich_click as click

from .console import console


def _get_module_name(path: Path, root: Path) -> str:
    path = path.with_suffix("")
    return ".".join(path.relative_to(root).parts)


@click.command(name="fuzz")
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
@click.option(
    "--process-count",
    "-n",
    default=(multiprocessing.cpu_count()),
    help="Number of processes to create for fuzzing.",
)
@click.option(
    "--coverage",
    type=int,
    is_flag=False,
    flag_value=-1,
    default=0,
    help="Number of processes to report coverage (0 = off).",
)
@click.option("--seed", "-s", "seeds", multiple=True, type=str, help="Random seeds")
@click.option(
    "--passive",
    is_flag=True,
    default=False,
    help="Print one process output into console, run other in background.",
)
@click.pass_context
def run_fuzz(
    ctx: click.Context,
    paths: Tuple[str],
    process_count: int,
    coverage: int,
    seeds: Tuple[str],
    passive: bool,
) -> None:
    """Run a Woke test using multiple processes."""
    import importlib.util
    import inspect
    import shutil

    from woke.config import WokeConfig
    from woke.testing.fuzzing.fuzzer import fuzz

    if coverage == -1:
        coverage = process_count

    if process_count < coverage:
        raise ValueError("Coverage must be less than or equal to process count.")

    config = WokeConfig()
    config.load_configs()  # load ~/.woke/config.toml and ./woke.toml

    random_seeds = [bytes.fromhex(seed) for seed in seeds]
    if len(paths) == 0:
        paths = (str(config.project_root_path / "tests"),)

    py_files = set()

    for path in paths:
        fuzz_path = Path(path).resolve()

        if fuzz_path.is_file() and fuzz_path.match("*.py"):
            py_files.add(fuzz_path)
        elif fuzz_path.is_dir():
            for p in fuzz_path.rglob("test_*.py"):
                if p.is_file():
                    py_files.add(p)
        else:
            raise ValueError(f"'{fuzz_path}' is not a Python file or directory.")

    fuzz_functions = []

    sys.path.insert(0, str(config.project_root_path))
    for file in py_files:
        module_name = _get_module_name(file, config.project_root_path)
        module_spec = importlib.util.spec_from_file_location(module_name, file)
        if module_spec is None or module_spec.loader is None:
            raise ValueError()
        module = importlib.util.module_from_spec(module_spec)
        sys.modules[module_name] = module
        module_spec.loader.exec_module(module)

        functions: Iterable[Tuple[str, Callable]] = (
            (func_name, func)
            for func_name, func in inspect.getmembers(module, inspect.isfunction)
            if func.__module__ == module_name and func_name.startswith("test")
        )
        for func_name, func in functions:
            console.print(f"Found '{func_name}' function in '{func.__module__}' file.")
            fuzz_functions.append((func_name, func))

    logs_dir = config.project_root_path / ".woke-logs" / "fuzz"
    shutil.rmtree(logs_dir, ignore_errors=True)
    logs_dir.mkdir(parents=True, exist_ok=True)

    try:
        for func_name, func in fuzz_functions:
            console.print("\n\n")
            console.print(f"Fuzzing '{func_name}' in '{func.__module__}'.")
            fuzz(
                config,
                func_name,
                func,
                process_count,
                random_seeds,
                logs_dir,
                passive,
                coverage,
                False,
            )
    except Exception as e:
        console.print_exception()
        sys.exit(1)
